from __future__ import absolute_import, division, print_function

from glue.viewers.common.qt.data_viewer import DataViewer

__all__ = ['DataViewerWithState']


class DataViewerWithState(DataViewer):
    pass
